require "kmCore"
local _dbg = require "coaf.debug"
local pt = require "pl.pretty"
local is_ok, FEATURES = pcall(require, "features")
local _M = {}
local codec

local function GetAesKeyLen( aesName )
	local aesKeyLen
	if aesName == "AES-128" or aesName == "aes-128" then
		aesKeyLen = 16
	elseif aesName == "AES-192" or aesName == "aes-192"  then
		aesKeyLen = 24
	elseif aesName == "AES-256" or aesName == "aes-256" then
		aesKeyLen = 32
	else
		aesKeyLen = 0
	end
	return aesKeyLen
end

local function onLookupRtspAuthUser(rtsp_server, digist, realm, username)
    if not username or username == "" then
		return nil
	end

    for i, v in ipairs(_G.media_users) do
        if v.username == username then
            return v.password
        end
    end
end

local function GetTsMuxrate(stream)
    local mode = _G.codec_cfg[stream].rc.mode
    if mode ~= "CBR" then
        return _G.codec_cfg[stream].rc.u32MaxBitRate * 1000 + 128000
    end

    return _G.codec_cfg[stream].rc.u32BitRate * 1000 + 128000
end

local function createNetPush(targetAddr, config, options, mediaSource, groupId, oldstream)
    local stream = oldstream
    if not stream then
        stream = KM_STREAMING.NetPusher.createNew("net-push:" .. groupId)
    end
    if not stream then
        _dbg.Error("STREAMS", "Cannot create Rtmp stream, Can not New NetPush Server")
        return
    end

    local protocol, orig_url_tail, s
	local user_in_url, pass_in_url
	local ttl_in_url, pkt_size_in_url

	protocol, s = targetAddr:match( "(%w+)://(.+)$" )
	if not protocol then
		_dbg.Error( "STREAMS", "Invalid target URL of:", targetAddr, "(Stream:", groupId )
		return nil
	end
	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
    	local s1 = s:match("(.-)/(.+)")
    	if s1 then
        	user_in_url = s1:match( "([^/]-)@./+$" )
    	end
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if protocol == "udp" or protocol == "rtp" then
		local pfx
		pfx, s = orig_url_tail:match( "(.-)%?(.+)$" )
		if pfx then
			orig_url_tail = pfx
			ttl_in_url = s:match("ttl=(%d+)" )
			ttl_in_url = tonumber(ttl_in_url)
			pkt_size_in_url = s:match("pkt_size=(%d+)")
			pkt_size_in_url = tonumber(pkt_size_in_url)
		end
	end

	if type(options) == "table" then
		local user = options["username"] or options["user"]
		local pass = options["password"] or options["pass"]
		if user and user ~= "" then user_in_url = user end
		if pass then pass_in_url = pass end
		if options["ttl"] then
			ttl_in_url = tonumber(options["ttl"]) or ttl_in_url
		end
		if options["pkt_size"] then
			pkt_size_in_url = tonumber(options["pkt_size"]) or pkt_size_in_url
		end
	else
		options = {
			muxrate = -1
		}
	end

	--Formal the URL
	targetAddr = protocol .. "://"
	if user_in_url then
		targetAddr = targetAddr .. user_in_url
		if pass_in_url then
			targetAddr = targetAddr .. ":" .. pass_in_url
		end
		targetAddr = targetAddr .. "@"
	end

	targetAddr = targetAddr .. orig_url_tail

	if protocol == "udp" or protocol == "rtp" then
		--if ttl_in_url or pkt_size_in_url then
			targetAddr = targetAddr .. "?"
			if ttl_in_url then
				targetAddr = targetAddr .. "ttl=" .. ttl_in_url
			else
				targetAddr = targetAddr .. "ttl=127"
			end
			targetAddr = targetAddr .. "&"
			if pkt_size_in_url and pkt_size_in_url > 0 and pkt_size_in_url <= 1472 then
				targetAddr = targetAddr .. "pkt_size=" .. pkt_size_in_url
			else
				targetAddr = targetAddr .. "pkt_size=1316"
			end
		--end
	end

    local sessionId = stream:AddSession(groupId, mediaSource, 4*1024*1024, 512*1024)
    if not sessionId then
        _dbg.Error("STREAMS", "Fail to add RTMP session for channel")
        return
    end

	if protocol:match("^rtmp") then --That may be: rtmp, rtmps, ...
		protocol = "rtmp"
	end

    options.PushPath = targetAddr
	--options.PushMode = "NONE" --What's means?
	options.PushProtocolType = protocol

    stream:SetOption( sessionId, options )

    return stream, sessionId, targetAddr
end

_M.defaultRtsp = {
    enable = true,
    session = "ch01",
    port = 554,
    httpTunnelPort = 8554,
    auth = false,
    no_adts = 0,
    ssm = 0,
    multicast_addr = "224.0.0.1",
    multicast_ttl = 127,
    multicast_video_port = 3101,
    multicast_audio_port = 3102,
}
if is_ok and FEATURES.FEATURE_DEFAULT_RTSP_AUTH then
    _M.defaultRtsp.auth = true
end

_M.defaultRtmp = {
    address = "192.168.0.111:8054",
    user = "admin",
    pass = "pass",
    connTimeout = 10,
    connIntv = 3,
    withMeta = true
}

_M.defaultSrt = {
    enable = true,
    address = "",
    connectionMode = "Listener",
    encryption = 0,
    passphrase = "",
    latency = 125,
    bandwidth = 5,
    payloadSize = 1316,
    listenerPort = 8083,
    ts_null_multiple = 0,
    ts_pcr_period = 20,
    ts_pts_pcr_delay = 500,
    ts_transport_stream_id = 101,
    srt_stream_id = ""
}

_M.defaultTs = {
    port = 9099,
    ts_pts_pcr_delay = 200,
    ts_service_name = "Encoder",
    ts_null_multiple = 1.5,
    ttl = 127,
    address = "192.168.0.224",
    ts_pcr_period = 20,
    ts_service_provider = "Encoder device",
    ts_transport_stream_id = 101,
    ts_start_pid = 481,
    ts_pmt_start_pid = 480,
    ts_tables_version = 6
}

_M.defaultNDIHX = {
    ttl = "1",
    NDIname = "Chan 1 name",
    group = "group",
    netmask = "255.255.0.0",
    connection = "tcp",
    netprefix = "239.255.0.0",
    discovery_server = "",
}

_M.defaultHLS = {
    session = "main2",
    segmentTime = 5000,
    maxSegments = 6,
    file_location = "/tmp/hls",
}

_M.defaultStreamCfg = {
    rtsp = _M.defaultRtsp,
    rtmp = _M.defaultRtmp,
    srt = _M.defaultSrt,
    ts = _M.defaultTs,
    hls = _M.defaultHLS,
    ndi_hx = _M.defaultNDIHX
}

local function createRtsp(cfg, mediaSource, _, oldstream)
    local config = table.copy(_M.defaultRtsp)
    table.update(config, cfg)

    local stream = oldstream
    if not stream then
        stream = KM_STREAMING.Rtsp.Server.createNew( 'KILOVIEW Rtsp Media Server', "*", config.port, config.httpTunnelPort or 8554 )
    end
    if not stream then
        _dbg.Error("Cannot create Rtsp Stream, Can not New Rtsp Server.")
        return
    end

    for i, v in ipairs(_G.media_users) do
        stream:AddAuth(v.username, v.password)
    end
    stream:EnableAuth(config.auth)

    local sessionId = stream:AddSession(config.session, mediaSource, {
        ssm = config.ssm,
		ttl = config.multicast_ttl,
		videoAddr = config.multicast_addr,
		audioAddr = config.multicast_addr,
		videoPort = config.multicast_video_port,
		audioPort = config.multicast_audio_port
    })
    if not sessionId then
        _dbg.Error( "STREAMS", "Fail to add RTSP session for channel" )
        return 
    end
    stream:SetDropADTSHeader(config.no_adts)
    return stream, sessionId
end

local function createNetRTMP(cfg, mediaSource, groupId, oldstream)
    local config = table.copy(_M.defaultRtmp)
    table.update(config, cfg)

    local options = {
        user = config.user,
        pass = config.pass,
        connTimeout = config.connTimeout,
        connIntv = config.connIntv,
        withMeta = config.withMeta,
        enableNtp = -1
    }

    local targetAddr = config.address
    return createNetPush(targetAddr, config, options, mediaSource, groupId, oldstream)
end

local function createNetSRT(cfg, mediaSource, groupId, oldstream, streamer)
    local config = table.copy(_M.defaultSrt)
    table.update(config, cfg)

    local srtAddress = config.connectionMode:upper() == "LISTENER" and "" or config.address
    local srtUrl = "srt://" .. srtAddress .. ":" .. config.listenerPort

    local muxrate = GetTsMuxrate(streamer.bindStream)
    local ts_null_multiple = config.ts_null_multiple
    if not tonumber(ts_null_multiple) or tonumber(ts_null_multiple) <= 1 then
		muxrate = -1
		ts_null_multiple = 1.0
    end

    local options = {
        PushPath = srtUrl,
        PushProtocolType = "srt",
        srt_connect_mode = config.connectionMode,
        srt_encryption_key_len = GetAesKeyLen(config.encryption),
        srt_encryption_key = config.passphrase,
        srt_latency = config.latency,									
        srt_bandwidth_overhead = config.bandwidth * 0.01,
        srt_payload_size = config.payloadSize,
        srt_stream_id = config.srt_stream_id,
        srt_max_bw = -1,
        srt_input_bw = 0,
        muxrate = muxrate,
        ts_null_multiple = ts_null_multiple,
        ts_pcr_period = config.ts_pcr_period,
        ts_pts_pcr_delay = config.ts_pts_pcr_delay,
        ts_transport_stream_id = config.ts_transport_stream_id
    }

    return createNetPush(srtUrl, config, options, mediaSource, groupId, oldstream)
end

local function createTs(cfg, mediaSource, groupId, oldstream, streamer)
    local config = table.copy(_M.defaultTs)
    table.update(config, cfg)

    local stream = oldstream
    if not stream then
        stream = KM_STREAMING.NetPusher.createNew( "udp-push:" .. groupId )
    end

    if not stream then
        _dbg.Error("Cannot create Ts stream, Can not New Ts Server")
        return
    end

    local sessionId = stream:AddSession("stream" .. groupId, mediaSource, {videoBuffer=8*1024*1024,audioBuffer=64*1024})
    if not sessionId then
        _dbg.Error("STREAMS", "Fail to add Ts session for channel")
        return
    end

    local muxrate = GetTsMuxrate(streamer.bindStream)
    local ts_null_multiple = config.ts_null_multiple
    if not tonumber(ts_null_multiple) or tonumber(ts_null_multiple) <= 1 then
		muxrate = -1
		ts_null_multiple = 1.0
    end

    local targetAddr = "udp://" .. config.address .. ":" .. config.port
    local PushPath = targetAddr .."?" .."ttl=".. config.ttl .."&pkt_size=1316"

    local options = {
        PushPath = PushPath,
        PushMode = "NONE",
        PushProtocolType = "udp",
        muxrate  = muxrate,
        ts_transport_stream_id =  config.ts_transport_stream_id,
        ts_pts_pcr_delay  = config.ts_pts_pcr_delay,
        ts_pmt_start_pid  = config.ts_pmt_start_pid,
        ts_start_pid  	  = config.ts_start_pid,
        ts_tables_version = config.ts_tables_version,
        ts_service_name   =  config.ts_service_name,
        ts_service_provider	=  config.ts_service_provider ,
        ts_null_multiple    =  ts_null_multiple ,
        ts_pcr_period       =  config.ts_pcr_period
    }

    stream:SetOption(sessionId, options)
    return stream, sessionId
end

local function createNDI_HX(cfg, mediaSource, groupId, oldstream, streamer)
    local config = table.copy(_M.defaultNDIHX)
    table.update(config, cfg)

    --NOTE: NDIService只需要创建一次就够了，后续只需要往里addsession和removesession就行了
    local stream = oldstream
    if not stream then
        stream = KM_STREAMING.ndiMediaSender.createNew()
    else
        stream:StopAllSessions()
		stream:RemoveAllSessions()
    end

    if not stream then
        _dbg.Info("Streams", "Can not create NDI service")
        return
    end

    table.merge(streamer.config, config)

    local serverIP
    if config.discovery_server and config.discovery_server ~= "" then
        local a, b, c, d = config.discovery_server:match("(%d+).(%d+).(%d+).(%d+)")
        a = tonumber(a)
        b = tonumber(b)
        c = tonumber(c)
        d = tonumber(d)
        if a and b and c and d and a > 0 and a < 256 and b > 0 and b < 256 and c > 0 and c < 256 and d > 0 and d < 256 then
            serverIP = config.discovery_server
        end
    end

    stream:SetOption{
        ttl = config.ttl,
        name = config.NDIname,
        group = config.group,
        netmask = config.netmask,
        connection = config.connection,
        netprefix = config.netprefix,
        discovery_server = serverIP,
    }

    stream:AddSession(0, mediaSource, {
        videoBuffer = 4*1024*1024, audioBuffer = 64*1024, fps = tostring(math.min(_G.capture_cfg[1].sourceFrameRate, _G.codec_cfg.main.rc.fr32DstFrmRate))
    })
    stream:AddSession(1, streamer.mediaSource_sub, {
        videoBuffer = 4*1024*1024, audioBuffer = 64*1024, fps = tostring(math.min(_G.capture_cfg[1].sourceFrameRate, _G.codec_cfg.sub.rc.fr32DstFrmRate))
    })
    return stream, tonumber(groupId:sub(#groupId))
end

local function createNewHLS(cfg, mediaSource, groupId, oldstream, streamer)
    local config = table.copy(_M.defaultHLS)
    table.update(config, cfg)

    local stream = oldstream
    if not stream then
        stream = KM_STREAMING.NetPusher.createNew( "hls-ts-fragment" )
    end

    if not stream then
        _dbg.Error( "STREAMS", "Fail to create TS fragment recorder group!" )
        return
    end

    local sessionId = stream:AddSession( "hls:" .. cfg.session, mediaSource, 4*1024*1024, 256*1024 )
    if not sessionId then
        _dbg.Error("STREAMS", "Fail to create HLS ts fragment session for Stream:", cfg.session)
        return
    end

    local location = config.file_location .. "/" .. cfg.session
    local pushSession = location .. "/playlist.m3u8"
    local segmentTime = math.ceil(config.segmentTime/1000) --To second(s)
    local segmentFilename = location .. "/SEG-%04d.ts"

    os.execute( "/bin/mkdir -p " .. location)
    os.execute(" /bin/rm -Rf " .. location .. "/*")

    stream:SetOption( sessionId, {
        PushPath = pushSession,
		PushProtocolType = "hls", --NOTE: The new version of moon_streamer should set type to 'hls'(not 'ts') to correct the timestamp problem.
		hls_time = segmentTime, --NOTE: This parameter should in 'second' unit.
		hls_list_size = config.maxSegments,
		hls_wrap = config.maxSegments,
		hls_segment_filename = segmentFilename,
    })
    return stream, sessionId
end

_M.createTypeStream = {
    rtsp = createRtsp,
    rtmp = createNetRTMP,
    SRT = createNetSRT,
    ts = createTs,
    ndi_hx = createNDI_HX,
    hls = createNewHLS,
}

function _M.createStream(cfg, mediaSource, ...)
    local func = _M.createTypeStream[cfg.type]
    return func(cfg, mediaSource, ...)
end

function _M.GetDefaultStreamConfig(cfg)
    local default = _M.defaultStreamCfg[cfg.type:lower()]
    if not default then
        return
    end
    local copy = table.copy(default)
    table.update(copy, cfg)
    return copy
end

return _M