local coaf = require 'coaf.core'
local thread = require 'coaf.thread'
local _dbg = require 'coaf.debug'
-- local utils = require "coaf.utils"
local hiperiphery = require 'hiperiphery'
local I2C = hiperiphery.I2C
local bit = require 'bit32'
local pl = require 'pl.pretty'
local Date = require 'pl.Date'
local TOOLS = require 'system.cmdtools'

pcall(require, 'lsystime')

--------
local _M = {
    i2c_dev = '/dev/i2c-0',
    BASE_ADDR = 0xA2, -- read A3h and write A2h
    i2c_handle = nil,
    dtfmt = 'yyyy-mm-dd HH:MM:SS'
}

function _M.ReadReg(addr, itemArr)
    if not _M.i2c_handle then
        return nil, 'Invalid I2C handle'
    end
    if not itemArr or type(itemArr) ~= 'table' then
        itemArr = {0xff, flags = I2C.I2C_M_RD}
    else
        if #itemArr < 1 then
            itemArr[1] = 0xff
        end
        if not itemArr.flags then
            itemArr.flags = I2C.I2C_M_RD
        else
            itemArr.flags = itemArr.flags | I2C.I2C_M_RD
        end
    end

    local addrs = {}
    local i
    for i = 1, #itemArr do
        addrs[i] = addr + i - 1
    end

    local msg = {addrs, itemArr}
    local done, err =
        pcall(
        function()
            _M.i2c_handle:transfer(_M.BASE_ADDR >> 1, msg)
        end
    )
    if not done then
        if type(err) == 'table' then
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
            return nil, err.message or '---'
        else
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
            return nil, err
        end
    else
        return itemArr
    end
end

-----------------------------------

function _M.WriteReg(addr, itemArr)
    if not _M.i2c_handle then
        return false, 'Invalid I2C handle'
    end
    if not itemArr or type(itemArr) ~= 'table' or #itemArr < 1 then
        return false, 'Invalid data'
    end

    --table.insert( itemArr, 1, addr )
    local addrs = {}
    local i
    for i = 1, #itemArr do
        addrs[i] = addr + i - 1
    end
    local msg = {addrs, itemArr}
    local done, err =
        pcall(
        function()
            _M.i2c_handle:transfer(_M.BASE_ADDR >> 1, msg)
        end
    )
    if not done then
        if type(err) == 'table' then
            --print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
            return nil, err.message or '---'
        else
            --print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
            return nil, err
        end
    else
        return true
    end
end

--------

function decToBcd(val)
    local d = string.format('%d', math.modf(val / 10))
    local d1 = tonumber(d * 10)
    local d2 = val - d1
    return tonumber(d * 16 + d2)
end

function bcdToDec(val)
    local hl = bit.rshift(val, 4)
    local hh = bit.band(val, 0xf)
    local hr = string.format('%d%d', hl, hh)
    return string.format('%d%d', hl, hh)
end

--------
function _M.Init()
    if LSYSTIME then
        LSYSTIME.SetAllowTimeDiff(100000) --100ms.
    end

    local done, err =
        pcall(
        function()
            _M.i2c_handle = I2C(_M.i2c_dev)
        end
    )
    if not done then
        if type(err) == 'table' then
            _dbg.Error('RTC', 'Could not open I2C device for RTC controlling. ErrMsg:', err.message or '---')
        else
            _dbg.Error('RTC', 'Could not open I2C device for RTC controlling. ErrMsg:', err)
        end
    else
        _dbg.Info('RTC', 'RTC controller openned, at ' .. string.format('%02x', _M.BASE_ADDR) .. '@' .. _M.i2c_dev)
    end
end
--------

function _M.GetTime()
    local wkd = {'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'}
    local datatime = _M.ReadReg(0x02, {0, 0, 0, 0, 0, 0, 0})
    if datatime ~= nil then
        -- pl.dump(datatime)
        local second = bcdToDec(bit.band(datatime[1], 0x7f))
        local minute = bcdToDec(bit.band(datatime[2], 0x7f))
        local hour = bcdToDec(bit.band(datatime[3], 0x3f))
        local day = bcdToDec(bit.band(datatime[4], 0x3f))
        -- local second = wkd[tonumber(bcdToDec(bit.band(datatime[5],0x7)))],
        -- local century_bit = bit.rshift( datatime[5],7)
        local weekday = bcdToDec(bit.band(datatime[5], 0x7))
        local month = bcdToDec(bit.band(datatime[6], 0x1f))
        local year = bcdToDec(datatime[7])
        if bit.band(datatime[6], 0x80) > 0 then
            year = '19' .. year
        else
            year = '20' .. year
        end

        local datetime_str = year .. '-' .. month .. '-' .. day .. ' ' .. hour .. ':' .. minute .. ':' .. second
        return datetime_str
    else
        _dbg.Error('RTC', 'GetTime error!!!')
        return nil
    end
end

-- print(_M.GetTime())

function _M.SetTime(datetime_str)
    if not datetime_str then
        _dbg.Error('RTC', 'no datatime string!')
        return false
    end

    local df = Date.Format(_M.dtfmt)
    local dt = df:parse(datetime_str)

    print(df:tostring(dt))
    if dt ~= nil and dt.tab then
        local century_month = 0x0
        local year = 0
        if dt.tab.year < 2000 then
            century_month = 0x80
            year = decToBcd(dt.tab.year - 1900)
        else
            century_month = 0x00
            year = decToBcd(dt.tab.year - 2000)
        end
        century_month = bit.bor(century_month, decToBcd(dt.tab.month))
        _M.WriteReg(
            0x02,
            {
                decToBcd(dt.tab.sec),
                decToBcd(dt.tab.min),
                decToBcd(dt.tab.hour),
                decToBcd(dt.tab.day),
                decToBcd(dt.tab.wday),
                century_month,
                year
            }
        )
    else
        _dbg.Error('RTC', 'SetTime() parse string error!!!')
        return false
    end
    return true
end
function _M.Store()
    local datetime = os.date('%Y-%m-%d %H:%M:%S', os.time())
    _dbg.Info('RTC', 'Set UTC RTC datetime=', datetime)
    return _M.SetTime(datetime)
end

function _M.Load()
    local datetime = _M.GetTime(datetime)
    _dbg.Info('RTC', 'Get UTC RTC datetime=', datetime)
    local df = Date.Format('yyyy-mm-dd HH:MM:SS')
    local dt = df:parse(datetime)

    if LSYSTIME then
        LSYSTIME.AdjustTime(datetime, -1000000, 1000000)
    else
        os.execute(TOOLS.SET_DATETIME .. " '" .. datetime .. "'")
    end
    return true
end

--------
return _M
