local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local client = require "coaf.client"
local adapter = require 'coaf.cli_adapter.simple'
local cjson = require "cjson"
local thread = require "coaf.thread"
local event = require "coaf.event"
local pt = require 'pl.pretty'
local websocket_server

local _M = {
    updating = {}, data = {}
}

client.InitSocket(nil, adapter)
local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		_dbg.Error( "CO", "Fail to create connection to service '" .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

-- 供给其他服务调用，更新websocket层的缓存数据
function _M.update_data(module, info)
    _M.data[module] = info
    websocket_server.send_msg(websocket_server.get_module_data({module}))
    return true
end

function _M.request_data(module)
    local is_wait
    while _M.updating[module] do
        is_wait = true
    end

    --NOTE: 这里就是确保只要请求了某个模块的数据，那就能拿到最新的数据并且不会做重复的coaf调用请求
    if is_wait then
        return
    end

    _M.updating[module] = true
    -- TODO: 向其他服务请求数据
    -- NOTE:这里建议用xpcall来运行，确保updating能正常修改为nil
    -- _M.data[module] = 
    _M.updating[module] = nil
    return
end

local cpu_mode = {"user", "nice", "system", "idle", "iowait", "irq", "softirq", "steal", "guest", "guest_nice"}
local cpu_info = {old = {}, new = {}}
local network_info = {}

local function getUnit(size, unit)
    unit = unit or "b"

    for _, v in ipairs({"kB", "MB", "GB"}) do
        if size > 1024 then
            unit = v
            size = size / 1024
        end
    end
    return string.format("%.2f%s", size, unit)
end

local function check_cpu()
    local cpu_stat = get_contents("/proc/stat")
    local cpu = {string.match(cpu_stat,
    "cpu  (%d+) (%d+) (%d+) (%d+) (%d+) (%d+) (%d+) (%d+) (%d+) (%d+)")}
    local diff = 0
    for i, v in ipairs(cpu_mode) do
        cpu_info.old[v] = cpu_info.new[v] or 0
        cpu_info.new[v] = cpu[i]
        diff = diff + (cpu_info.new[v] - cpu_info.old[v])
    end
    _M.data.cpu.precent = string.format("%.2f", 100 - (((cpu_info.new.idle or 0) - (cpu_info.old.idle or 0)) * 100) / diff)
end

local function check_mem()
    local meminfo = {}
    for line in io.lines("/proc/meminfo") do
        local name, size, unit = string.match(line, "([^:]+):%s+(%d+)%s?(k?B?)")
        if unit == 'kB' then
            size = size * 1024
        end
        meminfo[name] = tonumber(size)
    end
    _M.data.mem.total = getUnit(meminfo.MemTotal)
    _M.data.mem.used = getUnit(meminfo.MemTotal - (meminfo.MemFree + meminfo.Buffers + meminfo.Cached))
    _M.data.mem.precent = string.format("%.2f", ((meminfo.MemTotal - meminfo.MemFree - meminfo.Buffers - meminfo.Cached) / meminfo.MemTotal) * 100)
end

local function check_disk()
    local result = { total=0, used=0, free=0, percent=100 }
    local file = io.popen("/bin/df -m","r")
    if not file then
        return result
    end

    local line = file:read("l")
    while line and (result.total <= 0 ) do
        local flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+/$" )
        if flag then
            result.total = tonumber(t)
            result.used = tonumber(u)
            result.free = tonumber(f)
            result.percent = tonumber(p)
            break
        end
        line=file:read("l")
    end
    file:close()
    _M.data.disk.precent = string.format("%.2f", result.used / result.total * 100)
    _M.data.disk.total = result.total .. "MB"
    _M.data.disk.used = result.used .. "MB"
end

local function check_network()
    local cmd = "cat /proc/net/dev| grep eth0 | awk '{print $2}{print $10}'"
    local f = io.popen(cmd)
    if f then
        local rx = f:read("l") -- receive bytes
        local tx = f:read("l") -- transmit bytes
        f:close()
        network_info.last_rx = network_info.rx
        network_info.last_tx = network_info.tx

        network_info.tx = tx
        network_info.rx = rx
        _M.data.network.receive = (network_info.rx - (network_info.last_rx or 0)) * 8
        _M.data.network.transmit = (network_info.tx - (network_info.last_tx or 0)) * 8
    end

    websocket_server.send_msg(websocket_server.get_module_data({
        "cpu", "mem", "disk", "network", "bitrate"
    }))
end

local function check_RealBitrate()
    local proxy = GetProxy("/codec", "codec")
    if not proxy then
        return
    end

    local _, main_bitrate = proxy.GetRealBitrate("main")
    _M.data.bitrate[1] = _M.data.bitrate.main or main_bitrate

    local _, sub_bitrate = proxy.GetRealBitrate("sub")
    _M.data.bitrate[2] = _M.data.bitrate.sub or sub_bitrate

    proxy:Destroy()
end

local function onNsLookupNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.addresses = msg.addresses
end

local function NsLookup( targetHost )
	if not targetHost or targetHost == "" then
		return false, nil
	end

	local ctx = {
		status = "wait"
	}

	local topic = tostring(ctx)
	local ticket = _M.event:Subscribe( topic, onNsLookupNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/usr/bin/nslookup " .. targetHost )
		if not pfile then
			_M.event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local found_name
			local pfx, addr, addrs
			while line do
				if not found_name then
					found_name = line:match( "^[Nn]ame:%s+(.+)$" )
				else
					addr = line:match( "^[Aa]ddress.-:%s+(.+)$" )
					if addr then
						if not addrs then
							addrs = { addr }
						else
							table.insert( addrs, addr )
						end
					end
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				_M.event:NotifyBack( topic, { status="fail" } )
			else
				_M.event:NotifyBack( topic, { status="finished", code = code, addresses = addrs } )
			end
		end
		os.exit(0)
	end)

	local r,flag = coaf.WaitPID( pid )
	if not r then
		if flag == "#CoAF-KILL#" then
			os.execute( "/bin/kill " .. pid )
		end
		_M.event:Unsubscribe( ticket )
		return false, nil
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.1)	--Wait a while
		end
		_M.event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return ctx.code == 0, ctx.addresses
		else
			return false, nil
		end
	end
end

local address_tbl = {
    "baidu.com", "google.com", "kiloview.com"
}
local function check_Internet(thrObj)
    local skip
    for _, v in ipairs(address_tbl) do
        local done, result = NsLookup(v)
        if thrObj.terminated then 
			_M.data.network.status = "aborted"
			return
		end

        if done and result then
            _M.data.network.status = "internet-ready"
            return
        end
    end 
    _M.data.network.status = "dns-resolve-error"
end

function _M.Init()
    websocket_server = require "websocket_server"
    _M.data = websocket_server.data
    _M.data.cpu = {}
    _M.data.mem = {}
    _M.data.disk = {}
    _M.data.network = {}
    _M.data.bitrate = {}
    _M.event = event.createNew( "webSocket", true )
    --TODO: 定时获取本机 使用内存、储存空间、CPU负载、网络流量 的情况 
    thread.NewTimer(check_cpu, 1, 1):start() --CPU占用率
    thread.NewTimer(check_mem, 1, 1):start()
    thread.NewTimer(check_disk, 1, 10):start()
    thread.NewTimer(check_RealBitrate, 1, 1):start()
    thread.NewTimer(check_network, 1, 1):start()
    thread.NewTimer(function()
        local task = thread.New(check_Internet)
        if task then
            task:Start()
        end
    end, 1, 60):start()
end

return _M
