local coaf = require "coaf.core"
local thread = require "coaf.thread"
local event = require "coaf.event"
local _dbg = require "coaf.debug"

local compa = require "coaf.compatible"
local unistd = compa.CheckPosix( "posix.unistd" )
local errno = compa.CheckPosix( "posix.errno" )
local posix_time = compa.CheckPosix( "posix.time" )

require "coavahi"

--------
local _M = {
	REFRESH_PERIOD = 15,

	pubs = {},
	discos = {},

	discovered = {},
	discovery_callback = nil,
	avahi_daemon_running = nil,

	hostname = nil,
	set_hostname = true
}

--------

local function timediff( now, before )
	local diff = now.tv_sec - before.tv_sec
	diff = diff + (now.tv_nsec - before.tv_nsec)/1000000000
	return diff
end

--------
local function avahi_daemon_task( thrObj )
	local r
	local k, item, service, svcObj, resolvObj
	local i, event
	local all_events = {}
	local now,diff

	_dbg.Info( "AVAHI", "AVAHI daemon thread is running..." )
	_M.avahi_daemon_running = false
	while not thrObj.terminated do
		if not _M.avahi_daemon_running then
			_M.avahi_daemon_running = CoAF_AVAHI.Init()
			if not _M.avahi_daemon_running then
				_dbg.Error( "AVAHI", "Fail to initialize AVAHI client module" )
			else
				_dbg.Info( "AVAHI", "AVAHI client module is initialized." )
				_M.set_hostname = true
			end
		end
		if not _M.avahi_daemon_running then
			coaf.Sleep(1)
		else
			if _M.set_hostname and _M.hostname and CoAF_AVAHI.Ready() then
				if CoAF_AVAHI.SetHostname( _M.hostname ) then
					_dbg.Info( "AVAHI", "Hostname changed to:", _M.hostname )
					_M.set_hostname = nil
				end
			end

			for k, item in pairs( _M.pubs ) do
				if item.task and item.flag == "reset" then
					item:Destroy()
				end

				if not item.task and CoAF_AVAHI.Ready() then
					item:Create()
				end
				item.flag = nil
			end

			now = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )

			for k, item in pairs( _M.discos ) do
				if item.task and item.flag == "reset" then
					item:Destroy()
				end

				if not item.task and CoAF_AVAHI.Ready() then
					item:Create()
				end
				item.flag = nil

				if item.task then
					for i,service in pairs(item.services) do
						if not service.last_update_time or timediff(now, service.last_update_time) >= _M.REFRESH_PERIOD then
							--Force resolve the service
							if not item.task:Resolve( i, true ) then
								if service.resolve_info then
									service.resolve_info.untrust = true
								end
								table.insert( item.event_queue, {
									event = "resolve-failed",
									service = service
								})
							end
							service.last_update_time = now
						end
					end
				end

				if #item.event_queue > 0 then
					all_events[k] = item.event_queue
					item.event_queue = {}
				end
			end

			--Trigger the events.
			for k, item in pairs(all_events) do	
				if _M.discovery_callback then
					for i,event in ipairs(item) do
						_M.discovery_callback( k, event.event, event.service )
					end
				end
				all_events[k] = nil
			end

			r = CoAF_AVAHI.Step(0)
			coaf.Sleep(0.1)
			if not r then
				_dbg.Error( "AVAHI", "AVAHI client module run failed, try re-open!" )
				CoAF_AVAHI.Cleanup()
				_M.avahi_daemon_running = false
				coaf.Sleep(1) --Don't be too fast.
			end
		end
	end

	_M.StopAllPublishServices()
	_M.StopAllDiscovery()

	if _M.avahi_daemon_running then
		CoAF_AVAHI.Cleanup()
	end
	_dbg.Info( "AVAHI", "AVAHI daemon thread terminated." )
end

--------

function _M.Init()
end

--------
local pub_meta = {}
--------
local function onPubStatus( task, status, arg )
	local id = task:Name()
	_dbg.Info( "AVAHI", "Publish service of:", id, " status:", status, arg )
	local item = _M.pubs[id]
	if status == "cleanup" then
		if item then
			item.flag = "reset"
		end
	end
end
		
local function onPubError( task, err, msg )
	local id = task:Name()
	_dbg.Error( "AVAHI", "Publish service of:", id, " error - ", status, ":", arg )
	local item = _M.pubs[id]
	if item then
		if err == "collision" then --Name collision
			item.name_collision_suffix = item.name_collision_suffix + 1
		end

		item.flag = "reset"
	end
end

function pub_meta:Create()
	local name = self.name
	if self.name_collision_suffix > 0 then
		name = name .. "_" .. self.name_collision_suffix
	end
	self.task = CoAF_AVAHI.PublishService({
		name = name,
		service_type = self.service_type,
		port = self.port,
		txt = self.txts
	},{
		name = self.id, --This is the task-name.
		onError = onPubError,
		onStatus = onPubStatus
	})
	if not self.task then
		_dbg.Error( "AVAHI", "Fail to create publish task for:", self.name, "[", self.service_type, "]" )
		return false
	else
		_dbg.Info( "AVAHI", "Created publish task for:", self.name, "[", self.service_type, "]" )
		return true
	end
end

function pub_meta:Destroy()
	if self.task then
		self.task:Destroy()
		self.task = nil
		collectgarbage()
	end
end
--------
pub_meta.__index = pub_meta
--------

function _M.PublishService( id, name, service_type, port, txts )
	if not _M.internal_start_avahi() then
		return false, "Fail to start AVAHI"
	end

	if not id or not name or name == "" then
		return false, "Invalid arguments"
	end

	local item = _M.pubs[id]
	if item then
		_dbg.Info( "AVAHI", "Publish service:", id, " with name:", name, "service type:", service_type, ", but it's exist already! Try stop it then publish again." )
		item.flag = "reset"
	else
		item = {}
		setmetatable(item, pub_meta)
		_M.pubs[id] = item
	end

	item.id = id
	item.name = name
	item.name_collision_suffix = 0
	item.service_type = service_type or "_coaf._tcp"
	item.port = tonumber(port) or 14000

	if type(txts) ~= "table" then
		item.txts = { txts }
	else
		item.txts = txts
	end
	
	return true
end

--------
function _M.StopPublishService( id )
	if id and _M.pubs[id] then
		local item = _M.pubs[id]
		_M.pubs[id] = nil
		item:Destroy()
		return true
	else
		return false, "Not exist"
	end
end

function _M.StopAllPublishServices()
	local k,item
	for k,item in pairs(_M.pubs) do
		item:Destroy()
	end
	_M.pubs = {}
end

--------

function _M.internal_start_avahi()
	if not _M.avahi then
		_M.avahi = thread.New( avahi_daemon_task )
		_M.avahi:StartNow()
		return _M.avahi
	else
		return _M.avahi
	end
end

function _M.internal_destroy_avahi()
	if _M.avahi then
		_M.avahi:Terminate()
		_M.avahi = nil
		collectgarbage()
	end
end

----------------------------------------

local disco_meta = {}

local function onDiscoAdd( task, item_id, flag )
	local id = task:Name()
	local _self = _M.discos[ id ]
	if not _self then
		_dbg.Error( "AVAHI", "Invalid discovery task while 'onAdd' callback, for:", id )
		return
	end

	--_dbg.Info( "AVAHI", "Discovered [", flag, "] - ", item_id )

	if flag == "new" or flag == "update" then
		local svc = task:GetContentById( item_id )
		if not svc then
			_dbg.Error( "AVAHI", "Discovered item:", item_id, "but failed to get service object!" )
			return
		end

		local item_info = svc:ToTable()
		item_info.last_update_time = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
		local resolve_info = svc:GetResolveInfo()
		if resolve_info then
			item_info.resolve_info = resolve_info:ToTable()
		end
		_self.services[ item_id ] = item_info

		table.insert( _self.event_queue, {
			event = flag,
			service = item_info
		})

		if not task:Resolve( svc, true ) then --Force resolve.
			if item_info.resolve_info then
				item_info.resolve_info.untrust = true
			end
			table.insert( _self.event_queue, {
				event = "resolve-failed",
				service = item_info
			})
		end
	elseif flag == "resolve" then
		local svc = task:GetContentById( item_id )
		if not svc then
			_dbg.Error( "AVAHI", "Resolved item:", item_id, "but failed to get service object!" )
			return
		end
		local item_info = _self.services[item_id]
		if not item_info then
			_dbg.Error( "AVAHI", "Resolved item:", item_id, "but failed to get record!" )
			return
		end

		item_info.last_update_time = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
		local resolve_info = svc:GetResolveInfo()
		if resolve_info then
			item_info.resolve_info = resolve_info:ToTable()
		end

		table.insert( _self.event_queue, {
			event = "resolve",
			service = item_info
		})
	end
end

local function onDiscoRemove( task, item_id )
	local id = task:Name()
	local _self = _M.discos[ id ]
	if not _self then
		_dbg.Error( "AVAHI", "Invalid discovery task while 'onRemove' callback, for:", id )
		return
	end
	local item_info = _self.services[item_id]
	if item_info then
		_self.services[item_id] = nil
		table.insert( _self.event_queue, {
			event = "remove",
			service = item_info
		})
	end
end

local function onDiscoStatus( task, status, arg )
	local id = task:Name()
	local _self = _M.discos[ id ]
	if not _self then
		_dbg.Error( "AVAHI", "Invalid discovery task while 'onStatus' callback, for:", id )
		return
	end
	_dbg.Info( "AVAHI", "Discovery task:", id, "status:", status )
	if status == "cleanup" then
		_self.flag = "reset"
		_self.event_queue = {} --Clear all pending events
		_self.servcies = {} --Clear all.
		table.insert( _self.event_queue, {
			event = "cleanup"
		})
	end
	--TODO: other status.
end

local function onDiscoError( task, err, msg )
	local id = task:Name()
	local _self = _M.discos[ id ]
	if not _self then
		_dbg.Error( "AVAHI", "Invalid discovery task while 'onError' callback, for:", id )
		return
	end
	--Remove failed, so remove it.
	if err == "resolve-failed" then
		if msg ~= "invalid" then
			--NOTE: But must keep it in memory.
			--task:RemoveContentById( msg )
			local item_info = _self.services[msg]
			if item_info then
				if item_info.resolve_info then
					item_info.resolve_info.untrust = true
				end
				table.insert( _self.event_queue, {
					event = "resolve-failed",
					service = item_info
				})
			end
		end
	else
		_dbg.Info( "AVAHI", "Discovery task:", id, "error -", err, ":", msg )
		_self.flag = "reset"
	end
end

function disco_meta:Create()
	self.task = CoAF_AVAHI.BrowseService( self.service_type, nil, {
		name = self.service_type,
		onAdd = onDiscoAdd,
		onRemove = onDiscoRemove,
		onError = onDiscoError,
		onStatus = onDiscoStatus
	})
	if not self.task then
		_dbg.Error( "AVAHI", "Fail to create service discovery task for:", self.service_type )
		return false
	else
		_dbg.Info( "AVAHI", "Service discovery task for:", self.service_type, "is created." )
		return true
	end
end

function disco_meta:Destroy(no_trigger)
	if self.task then
		self.task:Destroy()
		self.task = nil
	end
	no_trigger = no_trigger or #self.services == 0
	self.event_queue = {} --Clear all events.
	self.services = {}
	if not no_trigger then
		table.insert( self.event_queue, {
			event = "cleanup"
		})
	end
end

disco_meta.__index = disco_meta

----------------------------------------

function _M.StartDiscovery(service_type)
	if not _M.internal_start_avahi() then
		return false, "Fail to start AVAHI"
	end
	service_type = service_type or "_coaf._tcp"
	local item = _M.discos[ service_type ]
	if item then
		item.flag = "reset"
	else
		item = {
			service_type = service_type,
			services = {},
			event_queue = {}
		}
		setmetatable(item, disco_meta)
		_M.discos[ service_type ] = item
	end
	return true
end

--------
function _M.StopDiscovery(service_type)
	service_type = service_type or "_coaf._tcp"
	local item = _M.discos[ service_type ]
	if item then
		_M.discos[ service_type ] = nil
		item:Destroy()
		return true
	else
		return false, "Not exists"
	end
end

function _M.StopAllDiscovery()
	local k, item
	for k,item in pairs(_M.discos) do
		item:Destroy(true)
	end
	_M.discos = {}
end

--------
function _M.SetDiscoveryCallback(cb)
	_M.discovery_callback = cb
end

--------
function _M.SetHostname(name)
	if not _M.internal_start_avahi() then
		return false, "Fail to start AVAHI"
	end
	_M.hostname = name
	_M.set_hostname = true
	return true
end
--------
return _M

